@@ -51,6 +51,16 @@
 
 #include "cd-drive.h"
 
+static struct {
+	const char *name;
+	gboolean can_write_cdr;
+	gboolean can_write_cdrw;
+	gboolean can_write_dvdr;
+	gboolean can_write_dvdram;
+} recorder_whitelist[] = {
+	{ "IOMEGA - CDRW9602EXT-B", TRUE, TRUE, FALSE, FALSE },
+};
+
 #if defined(__linux__) || defined(__FreeBSD__)
 
 /* For dvd_plus_rw_utils.cpp */
@@ -86,10 +96,10 @@ linux_bsd_media_type (const char *device)
 
 	fd = open (device, O_RDONLY);
 	if (fd < 0) {
-		return CD_MEDIA_TYPE_UNKNOWN;
+		return CD_MEDIA_TYPE_ERROR;
 	}
 
-	mmc_profile = get_mmc_profile ((void *)fd);
+	mmc_profile = get_mmc_profile ((void *)&fd);
 	close (fd);
 
 	switch (mmc_profile) {
@@ -186,7 +196,9 @@ struct scsi_unit {
 };
 
 struct cdrom_unit {
+	CDProtocolType protocol;
 	char *device;
+	char *display_name;
 	int speed;
 	gboolean can_write_cdr;
 	gboolean can_write_cdrw;
@@ -195,6 +207,32 @@ struct cdrom_unit {
 	gboolean can_read_dvd;
 };
 
+static char *cdrom_get_name (struct cdrom_unit *cdrom, struct scsi_unit *scsi_units, int n_scsi_units);
+
+static void
+add_whitelist (struct cdrom_unit *cdrom_s,
+	       struct scsi_unit *scsi_units, int n_scsi_units)
+{
+	guint i;
+
+	for (i = 0; i < G_N_ELEMENTS (recorder_whitelist); i++) {
+		if (cdrom_s->display_name == NULL) {
+			continue;
+		}
+
+		if (!strcmp (cdrom_s->display_name, recorder_whitelist[i].name)) {
+			cdrom_s->can_write_cdr =
+				recorder_whitelist[i].can_write_cdr;
+			cdrom_s->can_write_cdrw =
+				recorder_whitelist[i].can_write_cdrw;
+			cdrom_s->can_write_dvdr =
+				recorder_whitelist[i].can_write_dvdr;
+			cdrom_s->can_write_dvdram =
+				recorder_whitelist[i].can_write_dvdram;
+		}
+	}
+}
+
 static void
 get_scsi_units (char **device_str, char **devices, struct scsi_unit *scsi_units)
 {
@@ -314,7 +352,7 @@ get_device_max_speed (char *id)
 	const char *argv[20]; /* Shouldn't need more than 20 arguments */
 	char *dev_str, *stdout_data, *speed;
 
-	max_speed = 1;
+	max_speed = -1;
 
 	i = 0;
 	argv[i++] = "cdrecord";
@@ -383,11 +421,12 @@ cdrom_get_name (struct cdrom_unit *cdrom, struct scsi_unit *scsi_units, int n_sc
 		return NULL;
 	}
 	if (i == 2) {
+		g_free (cdrom->device);
 		cdrom->device = g_strdup(devfsname);
 	}
 	stdname[3] = '\0'; devfsname[14] = '\0'; /* just in case */
 	
-	if (stdname[0] == 's') {
+	if (cdrom->protocol == CD_PROTOCOL_SCSI) {
 		get_cd_scsi_id (cdrom->device, &bus, &id, &lun);
 		retval = get_scsi_cd_name (bus, id, lun, cdrom->device, scsi_units,
 			   n_scsi_units);
@@ -426,19 +465,20 @@ add_linux_cd_recorder (GList *cdroms,
 	cdrom = g_new0 (CDDrive, 1);
 
 	cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
+	cdrom->display_name = g_strdup (cdrom_s->display_name);
 
-	if (cdrom_s->device[0] == 's') {
-		cdrom->display_name = cdrom_get_name (cdrom_s, scsi_units,
-				n_scsi_units);
+	if (cdrom_s->protocol == CD_PROTOCOL_SCSI) {
+		cdrom->protocol = CD_PROTOCOL_SCSI;
 		if (!get_cd_scsi_id (cdrom_s->device, &bus, &id, &lun)) {
+			g_free (cdrom->display_name);
+			g_free (cdrom);
 			return cdroms;
 		}
 		cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d",
 				bus, id, lun);
 	} else {
+		cdrom->protocol = CD_PROTOCOL_IDE;
 		/* kernel >=2.5 can write cd w/o ide-scsi */
-		cdrom->display_name = cdrom_get_name (cdrom_s,
-				scsi_units, n_scsi_units);
 		cdrom->cdrecord_id = g_strdup_printf ("/dev/%s",
 				cdrom_s->device);
 	}
@@ -446,6 +486,9 @@ add_linux_cd_recorder (GList *cdroms,
 	if (recorder_only) {
 		cdrom->max_speed_write = get_device_max_speed
 			(cdrom->cdrecord_id);
+		if (cdrom->max_speed_write == -1) {
+			cdrom->max_speed_write = cdrom_s->speed;
+		}
 	} else {
 		/* Have a wild guess, the drive should actually correct us */
 		cdrom->max_speed_write = cdrom_s->speed;
@@ -469,10 +512,9 @@ add_linux_cd_recorder (GList *cdroms,
 	}
 	if (cdrom_s->can_read_dvd) {
 		cdrom->type |= CDDRIVE_TYPE_DVD_DRIVE;
+		add_dvd_plus (cdrom);
 	}
 
-	add_dvd_plus (cdrom);
-
 	return g_list_append (cdroms, cdrom);
 }
 
@@ -485,7 +527,7 @@ add_linux_cd_drive (GList *cdroms, struct cdrom_unit *cdrom_s,
 	cdrom = g_new0 (CDDrive, 1);
 	cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
 	cdrom->cdrecord_id = NULL;
-	cdrom->display_name = cdrom_get_name (cdrom_s, scsi_units, n_scsi_units);
+	cdrom->display_name = g_strdup (cdrom_s->display_name);
 	cdrom->device = g_strdup_printf ("/dev/%s", cdrom_s->device);
 	cdrom->max_speed_write = 0; /* Can't write */
 	cdrom->max_speed_read = cdrom_s->speed;
@@ -558,6 +600,8 @@ linux_scan (gboolean recorder_only)
 			*t = 0;
 		}
 		cdroms[j].device = get_cd_device_file (p);
+		/* Assume its an IDE device for now */
+		cdroms[j].protocol = CD_PROTOCOL_IDE;
 		if (t != NULL) {
 			p = t + 1;
 		}
@@ -570,6 +614,7 @@ linux_scan (gboolean recorder_only)
 	n_scsi_units = 0;
 	for (i = 0; i < n_cdroms; i++) {
 		if (cdroms[i].device[0] == 's') {
+			cdroms[i].protocol = CD_PROTOCOL_SCSI;
 			n_scsi_units++;
 		}
 	}
@@ -681,62 +726,180 @@ linux_scan (gboolean recorder_only)
 		maj = min = 0;
 	}
 	fclose(file);
-	
+
 	cdroms_list = NULL;
 	for (i = n_cdroms - 1, j = 0; i >= 0; i--, j++) {
+		if (have_devfs) {
+			char *s;
+			s = g_strdup_printf("%s cdroms/cdrom%d",
+					cdroms[i].device,  j);
+			g_free (cdroms[i].device);
+			cdroms[i].device = s;
+		}
+		cdroms[i].display_name = cdrom_get_name (&cdroms[i],
+				scsi_units, n_scsi_units);
+		add_whitelist (&cdroms[i], scsi_units, n_scsi_units);
+
 		if ((cdroms[i].can_write_cdr ||
 		    cdroms[i].can_write_cdrw ||
 		    cdroms[i].can_write_dvdr ||
 		    cdroms[i].can_write_dvdram) &&
-			(cdroms[i].device[0] == 's' ||
+			(cdroms[i].protocol == CD_PROTOCOL_SCSI ||
 			(maj > 2) || (maj == 2 && min >= 5))) {
-			if (have_devfs) {
-				cdroms[i].device = g_strdup_printf("%s cdroms/cdrom%d",
-						cdroms[i].device,  j);
-			}
 			cdroms_list = add_linux_cd_recorder (cdroms_list,
 					recorder_only, &cdroms[i],
 					scsi_units, n_scsi_units);
 		} else if (!recorder_only) {
-			if (have_devfs) {
-				cdroms[i].device = g_strdup_printf("%s cdroms/cdrom%d",
-						cdroms[i].device,  j);
-			}
 			cdroms_list = add_linux_cd_drive (cdroms_list,
 					&cdroms[i], scsi_units, n_scsi_units);
 		}
 	}
 
-	g_free (scsi_units);
+	for (i = n_cdroms - 1; i >= 0; i--) {
+		g_free (cdroms[i].display_name);
+		g_free (cdroms[i].device);
+	}
 	g_free (cdroms);
 
+	for (i = n_scsi_units - 1; i >= 0; i--) {
+		g_free (scsi_units[i].vendor);
+		g_free (scsi_units[i].model);
+		g_free (scsi_units[i].rev);
+	}
+	g_free (scsi_units);
+
 	return cdroms_list;
 }
 #endif /* __linux__ */
 
 #ifdef __FreeBSD__
+static void
+get_cd_properties (char *id, int *max_rd_speed, int *max_wr_speed,
+	CDDriveType *type)
+{
+    	int i;
+	const char *argv[20];
+	char *dev_str, *stdout_data, *rd_speed, *wr_speed, *drive_cap;
+
+	*max_rd_speed = -1;
+	*max_wr_speed = -1;
+	*type = 0;
+
+	i = 0;
+	argv[i++] = "cdrecord";
+	argv[i++] = "-prcap";
+	dev_str = g_strdup_printf ("dev=%s", id);
+	argv[i++] = dev_str;
+	argv[i++] = NULL;
+
+	if (g_spawn_sync (NULL,
+			  (char **)argv,
+			  NULL,
+			  G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL,
+			  NULL, NULL,
+			  &stdout_data,
+			  NULL,
+			  NULL,
+			  NULL)) {
+		wr_speed = strstr (stdout_data, "Maximum write speed:");
+		if (wr_speed != NULL) {
+		    	char *tok;
+		    	wr_speed += strlen ("Maximum write speed:");
+			for (tok = strtok (wr_speed, " (),\t\n");
+				tok && strcmp (tok, "CD");
+				tok = strtok (NULL, " (),\t\n")) {}
+			tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
+			*max_wr_speed = atol (tok);
+		}
+		else {
+		    	wr_speed = strstr (stdout_data, "Maximum write speed in kB/s:");
+			if (wr_speed != NULL) {
+			    	char *tok;
+			    	wr_speed += strlen ("Maximum write speed in kB/s:");
+				for (tok = strtok (wr_speed, " (),\t\n");
+					tok && strcmp (tok, "CD");
+					tok = strtok (NULL, " (),\t\n")) {}
+				tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
+				*max_wr_speed = atol (tok);
+			}
+		}
+
+		rd_speed = strstr (stdout_data, "Maximum read  speed:");
+		if (rd_speed != NULL) {
+		    	char *tok;
+			rd_speed += strlen ("Maximum read  speed:");
+			for (tok = strtok (rd_speed, " (),\t\n"); 
+				tok && strcmp (tok, "CD"); 
+				tok = strtok (NULL, " (),\t\n")) {}
+			tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
+			*max_rd_speed = atol (tok);
+		}
+		else {
+		    	rd_speed = strstr (stdout_data, "Maximum read  speed in kB/s:");
+			if (rd_speed != NULL) {
+			    	char *tok;
+			    	rd_speed += strlen ("Maximum read  speed in kB/s:");
+				for (tok = strtok (rd_speed, " (),\t\n");
+					tok && strcmp (tok, "CD");
+					tok = strtok (NULL, " (),\t\n")) {}
+				tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
+				*max_rd_speed = atol (tok);
+			}
+		}
+		drive_cap = strstr (stdout_data, "Does write DVD-RAM media");
+		if (drive_cap != NULL) {
+		    	*type |= CDDRIVE_TYPE_DVD_RAM_RECORDER;
+		}
+		drive_cap = strstr (stdout_data, "Does read DVD-R media");
+		if (drive_cap != NULL) {
+		    	*type |= CDDRIVE_TYPE_DVD_RW_RECORDER;
+		}
+		drive_cap = strstr (stdout_data, "Does read DVD-ROM media");
+		if (drive_cap != NULL) {
+		    	*type |= CDDRIVE_TYPE_DVD_DRIVE;
+		}
+		drive_cap = strstr (stdout_data, "Does write CD-RW media");
+		if (drive_cap != NULL) {
+		    	*type |= CDDRIVE_TYPE_CDRW_RECORDER;
+		}
+		drive_cap = strstr (stdout_data, "Does write CD-R media");
+		if (drive_cap != NULL) {
+		    	*type |= CDDRIVE_TYPE_CD_RECORDER;
+		}
+		drive_cap = strstr (stdout_data, "Does read CD-R media");
+		if (drive_cap != NULL) {
+		    	*type |= CDDRIVE_TYPE_CD_DRIVE;
+		}
+		g_free (stdout_data);
+	}
+
+	g_free (dev_str);
+}
+
 static GList *
 freebsd_scan (gboolean recorder_only)
 {
 	GList *cdroms_list = NULL;
 	const char *dev_type = "cd";
-	int fd;
 	int speed = 16; /* XXX Hardcode the write speed for now. */
-	int max_speed = CDR_MAX_SPEED;
 	int i = 0;
+	int cnode = 1; /* Use the CD device's 'c' node. */
 
 	while (1) {
-		CDDriveType type;
 		CDDrive *cdrom;
-		gchar *cam_path, *dev;
+		gchar *cam_path;
 		struct cam_device *cam_dev;
 
-		type = 0;
 		cam_path = g_strdup_printf ("/dev/%s%dc", dev_type, i);
 
 		if (!g_file_test (cam_path, G_FILE_TEST_EXISTS)) {
 			g_free (cam_path);
-			break;
+			cam_path = g_strdup_printf ("/dev/%s%d", dev_type, i);
+			cnode = 0;
+			if (!g_file_test (cam_path, G_FILE_TEST_EXISTS)) {
+				g_free (cam_path);
+				break;
+			}
 		}
 
 		if ((cam_dev = cam_open_spec_device (dev_type, i, O_RDWR, NULL)) == NULL) {
@@ -745,47 +908,41 @@ freebsd_scan (gboolean recorder_only)
 			continue;
 		}
 
-		/* XXX Other controllers might need to be added. */
-		if ((strncmp (cam_dev->sim_name, "ata", 3)) == 0) {
-			dev = g_strdup_printf ("/dev/a%s%dc", dev_type, i);
-		} else {
-			dev = g_strdup (cam_path);
-		}
+		cdrom = g_new0 (CDDrive, 1);
+		cdrom->display_name = g_strdup_printf ("%s %s", cam_dev->inq_data.vendor, cam_dev->inq_data.revision);
+		cdrom->device = g_strdup (cam_path);
+		cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d", cam_dev->path_id, cam_dev->target_id, cam_dev->target_lun);
+		/* Attempt to get more specific information from
+		 * this drive by using cdrecord.
+		 */
+		get_cd_properties (cdrom->cdrecord_id,
+			&(cdrom->max_speed_read),
+			&(cdrom->max_speed_write),
+			&(cdrom->type));
+		if (cdrom->type & CDDRIVE_TYPE_CD_RECORDER
+				|| cdrom->type & CDDRIVE_TYPE_CDRW_RECORDER
+				|| cdrom->type & CDDRIVE_TYPE_DVD_RAM_RECORDER
+				|| cdrom->type & CDDRIVE_TYPE_DVD_RW_RECORDER
+				|| !recorder_only) {
 
-		g_free (cam_path);
+			if (cdrom->max_speed_read == -1) {
+		    		cdrom->max_speed_read = speed;
+			}
+			if (cdrom->max_speed_write == -1) {
+			    	cdrom->max_speed_write = speed;
+			}
 
-		if ((fd = open (dev, O_RDWR, 0)) < 0) {
-			g_free (dev);
-			free (cam_dev);
-			i++;
-			continue;
-		}
-
-		type |= CDDRIVE_TYPE_CD_DRIVE;
-		if (ioctl (fd, CDRIOCWRITESPEED, &max_speed) >= 0) {
-			/* XXX is it CD-R or CD-RW ? */
-			type |= CDDRIVE_TYPE_CD_RECORDER;
-		}
-
-		close (fd);
-
-		if (type & CDDRIVE_TYPE_CD_RECORDER
-				|| type & CDDRIVE_TYPE_CDRW_RECORDER
-				|| recorder_only == FALSE) {
-			cdrom = g_new0 (CDDrive, 1);
-			cdrom->display_name = g_strdup_printf ("%s %s", cam_dev->inq_data.vendor, cam_dev->inq_data.revision);
-			cdrom->device = g_strdup (dev);
-			cdrom->max_speed_read = speed;
-			cdrom->max_speed_write = speed;
-			cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d", cam_dev->path_id, cam_dev->target_id, cam_dev->target_lun);
-			cdrom->type = type;
-
-			add_dvd_plus (cdrom);
+			if (cdrom->type & CDDRIVE_TYPE_DVD_DRIVE) {
+				add_dvd_plus (cdrom);
+			}
 
 			cdroms_list = g_list_append (cdroms_list, cdrom);
 		}
+		else {
+		    	cd_drive_free (cdrom);
+		}
 
-		g_free (dev);
+		g_free (cam_path);
 		free (cam_dev);
 
 		i++;
@@ -831,6 +988,7 @@ cd_drive_free (CDDrive *drive)
 	g_free (drive->display_name);
 	g_free (drive->cdrecord_id);
 	g_free (drive->device);
+	g_free (drive);
 }
 
 CDMediaType
@@ -839,7 +997,7 @@ guess_media_type (const char *device_path)
 #if defined(__linux__) || defined(__FreeBSD__)
 	return linux_bsd_media_type (device_path);
 #else
-	return CD_BLANK_MEDIA_TYPE_UNKNOWN;
+	return CD_MEDIA_TYPE_UNKNOWN;
 #endif
 }
 
